import openai, os

from .manual.functions import makeFunctions
from .methods.chat import chat
from .methods.embed import embed
from .methods.training import trainFile
from .methods.route import route

class Router:
    def __init__(self):
        self.buckets = {}
        pass

    def trainWithFile(self, file, batchSize):
        instructions = makeFunctions(self.buckets)
        code = trainFile(file, instructions, chat, embed, batchSize)
        return code[0], code[1]
    
    def addBuckets(self, buckets):
        self.buckets.update(buckets)
        return 200
    
    def removeBuckets(self, buckets):
        if buckets in self.buckets:
            return 404
        else:
            self.buckets.remove(buckets)
            return 200
    
    def route(self, question):
        return route(question, embed)

    def manualAnswer(self, question):
        return chat(question, makeFunctions(self.buckets), False), 200

    def learnAndRoute(self, question):
        instructions = makeFunctions(self.buckets)
        code = chat(question, instructions, True), 200
        return code

    def wipe(self):
        with open(r"router\db\trainingData.json", 'w') as file:
            pass
        return 200



import openai
import json
from .embed import embed

import threading
lock = threading.Lock()

def chat(prompt, funcs, write, instructions="Classify this item."):
    output = openai.ChatCompletion.create(
        model="gpt-3.5-turbo-16k-0613",
        messages=[
            {"role": "system", "content": "Always answer in a function call."},
            {"role": "system", "content": instructions},
            {"role": "user", "content": prompt},
        ],
        functions=funcs,
        function_call="auto",
    )

    function_call_name = output["choices"][0]["message"]["function_call"]["name"]

    data_to_save = {"prompt": embed(prompt), "output": function_call_name}

    if write == True:
        with lock:  # Acquire the lock before reading the file
            try:
                with open("router//db//trainingData.json", "r") as json_file:
                    try:
                        data = json.load(json_file)
                    except json.decoder.JSONDecodeError:
                        data = []
                    data.append(data_to_save)

            except FileNotFoundError:
                data = [data_to_save]

        with lock:  # Acquire the lock before writing to the file
            with open("router//db//trainingData.json", "w") as json_file:
                json.dump(data, json_file)

        # print(prompt + " -> " + function_call_name)

    return function_call_name


import math, numpy as np
from numpy.linalg import norm

def cos(vector1, vector2):

    vector1 = np.array(vector1)
    vector2 = np.array(vector2)

    cosine = np.dot(vector1,vector2)/(norm(vector1)*norm(vector2))  
    
    return cosine


import openai, json
from .cos import cos

def route(question, embed):
    auxillary = {}

    with open(r"router\db\trainingData.json", "r") as json_file:
            try:
                data = json.load(json_file)
            except json.decoder.JSONDecodeError:
                print("[NO DATA]")
    
    query = embed(question)
    
    for question in data:
        similarity = cos(query, question["prompt"])
        auxillary.update({similarity : question["output"]})

    auxillary = dict(sorted(auxillary.items(), reverse=True))
    # print(auxillary)
    auxillary = list(auxillary.values())
    return auxillary[0], 200


error = {
        "name": "error",
        "description": "Use this when you cannot classify the input with your options.",
        "parameters": {
            "type": "object",
            "properties": {
                "reason": {
                    "type": "string",
                    "description": "In 2 sentences, explain why you cannot classify it.",
                }
            },
            "required": ["reason"],
        },
    }


import openai, time, os
from TearRouting.router.inner import Router

class Tear:
    def __init__(self, api_key, organization = False):
        openai.api_key = api_key
        if organization: openai.organization = org
        self.internalRouter = Router()
        self.buckets = self.internalRouter.buckets
    
    def addBuckets(self, buckets): # Add training categories.
        code = self.internalRouter.addBuckets(buckets)
        return code

    def removeBuckets(self, buckets): # Remove training categories.
        code = self.internalRouter.removeBuckets(buckets)
        return code
    
    def train(self, questions, batchSize=5): # Train the model from a list of questions.
        code = self.internalRouter.trainWithFile(questions, batchSize)
        return {"code" : code[0], "cost" : code[1]}

    def manualRoute(self, inp): # Route the question with LLM instead of embeddings without adding to training data.
        code = self.internalRouter.manualAnswer(inp)
        return {"output" : code[0], "code" : code[1]}

    def route(self, inp): # Route the question using embeddings.
        code = self.internalRouter.route(inp)
        return {"output" : code[0], "code" : code[1]}

    def routeLearn(self, inp): # Train the model while also routing questions.
        code = self.internalRouter.learnAndRoute(inp)
        return {"output" : code[0], "code" : code[1]}

    def wipe(self): # Wipe the training data.
        code = self.internalRouter.wipe()
        return code


import openai

def embed(inp):
    response = openai.Embedding.create(
        input=inp,
        model="text-embedding-ada-002"
    )
    embeddings = response['data'][0]['embedding']
    return embeddings


from .error import error

def makeFunctions(buckets):
    functions = []
    functions.append(error)
    for bucket in buckets:
        temp = {
            "name": bucket.replace(" ", "_"),
            "description": buckets[bucket],
            "parameters": {
                "type": "object",
                "properties": {
                    "reason": {
                        "type": "string",
                        "description": f"The reason you classified the input as a {bucket}",
                    }
                },
                "required": ["reason"],
            },
        }
        functions.append(temp)
    
    return functions



import threading
import json, random, time

def trainFile(file, instructions, chatFunction, embedFunction, batch_size):
    iterator = 0
    cost = 0

    while iterator < len(file):
        threads = []

        for i in range(iterator, min(iterator + batch_size, len(file))):
            thread = threading.Thread(target=chatFunction, args=(file[i], instructions, True))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        smallCost = round(batch_size*random.randint(290,310)*(0.002/1000),5)
        cost += smallCost

        print(f"Epoch {1 + iterator//batch_size}/{(len(file) + batch_size - 1)//batch_size} ~ {smallCost}$")
        iterator += batch_size
        code = 200

    return code, cost
    



